// File   : C:\cbproj\SoundUtl\SevenBitPackets.c
// Date   : 2023-07-30
// Author : Wolfgang Buescher (DL4YHF)
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Copyright, License, and Disclaimer :
//      Copyright (c) 2023, Wolfgang Bscher (DL4YHF) .
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are met:
//   1. Redistributions of source code must retain the above copyright notice,
//      this list of conditions and the following disclaimer.
//   2. Redistributions in binary form must reproduce the above copyright notice,
//      this list of conditions and the following disclaimer in the
//      documentation and/or other materials provided with the distribution.
//   3. Neither the name of the author nor the name of other contributors
//      may be used to endorse or promote products derived from this software
//      without specific prior written permission.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
//  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
//   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
//   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
//   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
// STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
// EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//
// Purpose: 'Seven-bit packet encoder / decoder for audio via serial port' .
//     A relatively simple, but versatile / extensible and fairly robust method
//     to exchange multi-channel audio samples via the SERIAL PORT
//     (aka "COM port" under windows) .
//   * Uses only EIGHT BITS per 'serial frame' (8-N-1 format),
//          where bit SEVEN is exclusively used to mark the PACKET HEADER BYTE.
//   * Any other byte ("optional / extra header byte" or "payload")
//           has bit 7 (= the MSBit) *cleared* .
//   * Supports multiple data types, 'bit depths', and number of channels
//   * Details / specification of the basic packet format not HERE but in
//          SpecLab/html/Seven_Bit_Packets_for_the_serial_port.htm .
//
//  Module Prefix (to avoid "C" namespace pollution) :
//     "SBP_" = "Seven-Bit Packets" .
//
// Prerequisites / Dependencies :
//     "As few as possible", with the intention of using
//     this module also in certain microcontroller projects,
//     e.g. for dsPIC / PIC24 / ARM Cortex-M .
//     (The target CPU must support efficient DATA POINTERS,
//      including "pointers to structure", thus no chance
//      to use this module with crappy old 8-bit PICs
//      with its bizarre "FSR"- and "INDF"-register,
//      and the complete lack of 'modern' addressing modes
//      like "index register plus address offset").
//     This code also won't run on BIG ENDIAN machines,
//     because these days (except for Internet protocols)
//     BIG ENDIAN (aka "most significant byte first)
//     is fortunately out of fashion), and the bit-fiddling
//     requred to map larger integers into the seven-bit
//     payload is already tough enough !
//
//  When compiled as part of Spectrum Lab,
//     this module (SevenBitPackets) detects the C compiler
//     by the macro definition of __BORLANDC__ .
//     In that case ("compiling to be used in Spectrum Lab"),
//     SevenBitPackets.h #defines SBP_COMPILING_SPECLAB to ONE .
//
//  Symbols written in ALL UPPER CASE, possibly with underscores,
//      are either macro constants, or refer to the 'Seven Bit Packet'
//      specification, or both (names of important macro constants are the same
//      in SevenBitPackets.c and in Seven_Bit_Packets_for_the_serial_port.htm) .
//

#include "SevenBitPackets.h"  // header for THIS module ('Seven-bit packet encoder / decoder for audio via serial port')

#include <string.h> // for some reason, memset() is in string.h !
#include <math.h>   // using fmod() for our 'truncatable' fixed-point numbers (with OPTIONAL fractional part).
                    // Not required in a microcontroller firmware if the sampling rate is always an INTEGER .

#if( SBP_BUILT_IN_UNIT_TESTS )
# include <stdlib.h>        // rand(), used for the built-in unit test
# if( SBP_COMPILING_SPECLAB )
#  include "Utility1.h"  // stuff like DEBUG_EnterErrorHistory, LastErrorCodeToString, ..
# endif // SBP_COMPILING_SPECLAB ?
#endif // SBP_BUILT_IN_UNIT_TESTS ?


//---------------------------------------------------------------------------
// Internal constants and macros
//  (don't need a module prefix; no risk of "namespace pollution")
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Simple subroutines
//    (used for packet assembly / disassembly, avoids duplicating code)
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
int SBP_GetHeaderSizeInBytes( // determines the NUMBER of header bytes (1..4) from the PACKET HEADER BYTE
        BYTE bPacketHeaderByte ) // [in] packet header byte (the one with SB_PACKET_HEADER_FLAG *set*)
{
  int nHeaderBytes = 1; // at least ONE byte : the 'packet header byte' itself !
  if( (bPacketHeaderByte & SB_PACKET_HDR_LENGTH_MASK ) == SB_PACKET_HDR_LENGTH_LONG_PAYLOAD )
   { // 'long' payload length in two subsequent bytes ->
     nHeaderBytes += 2;
   }
  if( (bPacketHeaderByte & SB_PACKET_HDR_TYPE_MASK ) == SB_PACKET_HDR_TYPE_OTHER )
   { // PACKET_CONTENT_TYPE specified in another ("extra") header byte ->
     nHeaderBytes += 1;
   }
  return nHeaderBytes;
} // end SBP_GetHeaderSizeInBytes()




//---------------------------------------------------------------------------
// Functions operating on the T_SBP_Instance struct .
//  In C++, T_SBP_Instance would be a "class", and these would be "methods".
//  But forget about C++; this code was intended to run on small uCs.
//---------------------------------------------------------------------------



//---------------------------------------------------------------------------
CPROT void SBP_InitInstance( T_SBP_Instance *pInst )
  // Initializes all members of struct 'T_SBP_Instance', including the packet buffer.
{
  if( pInst != NULL )
   { memset( pInst, 0, sizeof(T_SBP_Instance) );

     // What about 'float' and 'double' members in this struct after 'memset zero' ?
     // Someone said :
     // > When all bits (sign, exponent, mantissa) are zero the floating point
     // > value represents also zero, as defined by IEEE 754. More specifically,
     // > that value is a "positive zero", also written as +0.

     // Prevent pointer chaos - let the source- and destination pointer for
     // the "bit-fiddling" functions point into our own 'packet buffer' :
     pInst->pbSrc = pInst->pbDst = pInst->bPacketBuffer;

   }
} // end SBP_InitInstance()



//---------------------------------------------------------------------------
// Functions that we may need in the uC firmware AND in Spectrum Lab :
//    * Extraction of longer integers (1 .. 32 bits)
//            from DENSELY PACKED(*) seven-bit payload
//    * Packing of longer integers (1 .. 32 bits)
//            into DENSELY PACKED(*) seven-bit payload
//   (*) DENSELY PACKED means there are no 'unused bits' between
//       individual bitgroups (with 1 to 32 bits each) in the
//       7-bit payload fields, in each byte of the 'packet buffer'.
//---------------------------------------------------------------------------

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Instead of a simple 'byte pointer' or 'byte index' into the payload buffer,
// the following functions also need to count the number of BITS in the
// packet buffer (T_SBP_Instance.bPacketBuffer[]) .
// If samples from an ADC or for a DAC don't "travel" with 7, 14, 21, or 28
// bits in our 'Seven-Bit-Payload-Packets', and more than a single channel
// shall be sent/received in a DENSELY PACKED packet, we need the following
// 'bit fiddling' functions SBP_WriteBitgroup()
// and SBP_GetBitgroupFromBuffer_Unsigned() / SBP_GetBitgroupFromBuffer_Unsigned().
// All of these 'bit fiddling' functions use 'packet buffer' as in- or output:
//   T_SBP_Instance.bPacketBuffer[SBP_PACKET_BUFFER_SIZE] : source or destination in RAM
//   T_SBP_Instance.iReadBitIndex : BIT- (not BYTE-) index for
//                   'appending to' / 'reading from' bPacketBuffer[]
//
// See "Example with 'densely packed' 24-bit samples, two channels" in
//  SpecLab/html/Seven_Bit_Packets_for_the_serial_port.htm#densely_packed_audio_sample :
//
// "Bit layout" in the payload, with BITS IN A BYTE shown in reversed order for clarity.
//  The PACKET_HEADER_BYTE (in packet byte index 0) is omitted for space reasons.
//
//  PacketBufferBitIndex (with tens in one line, and ones in the next) :
//    0 0 1 1 1 1 1 1   1 1 1 1 2 2 2 2   2 2 2 2 2 2 3 3   3 3 3 3 3 3 3 3   4 4 4 4 4 4 4 4   4 4 5 5 5 5 5 5   5 5 5 5 6 6 6 6
//    8 9 0 1 2 3 4 5   6 7 8 9 0 1 2 3   4 5 6 7 8 9 0 1   2 3 4 5 6 7 8 9   0 1 2 3 4 5 6 7   8 9 0 1 2 3 4 5   6 7 8 9 0 1 2 3
//
//   -Packet Byte [1]- -Packet Byte [2]- -Packet Byte [3]- -Packet Byte [4]- -Packet Byte [5]- -Packet Byte [6]- -Packet Byte [7]- .
//   [0 1 2 3 4 5 6 7] [0 1 2 3 4 5 6 7] [0 1 2 3 4 5 6 7] [0 1 2 3 4 5 6 7] [0 1 2 3 4 5 6 7] [0 1 2 3 4 5 6 7] [0 1 2 3 4 5 6 7] .
//    : : : : : : : 0(bit 7 cleared)  0   : : : : : : : 0   : : : : : : : 0   : : : : : : : 0   : : : : : : : 0   : : : : : : : 0  .
// Bit numbers in channels 'A' and 'B'    : : : : : : :     : : : : : : :     : : : : : : :     : : : : : : :     : : : : : : :/|\ .
//    : : : : : : :     : : : : : : :     : : : : : : :     : : : : : : :     : : : : : : :     : : : : : : :     : : : : : : : |  .
//    A A A A A A A     A A A A A A A     A A A A A A A     A A A B B B B     B B B B B B B     B B B B B B B     B B B B B B - |  .
//    0 0 0 0 0 0 0     0 0 0 1 1 1 1     1 1 1 1 1 1 2     2 2 2 0 0 0 0     0 0 0 0 0 0 0     1 1 1 1 1 1 1     1 1 2 2 2 2 - |  .
//    0 1 2 3 4 5 6     7 8 9 0 1 2 3     4 5 6 7 8 9 0     1 2 3 0 1 2 3     4 5 6 7 8 9 0     1 2 3 4 5 6 7     8 9 0 1 2 3 - |  .
//    |<------------------ Channel A -------------------------->| |<------------------- Channel B ------------------------->| ^ |  .
//                  ^                 ^                 ^                 ^                                                  /|\|  .
//                 /|\               /|\               /|\               /|\             One unused ("remaining") bit HERE ---' |  .
//    ,-------------'-----------------'-----------------'-----------------'-----------------------------------------------------'  .
//    '-- Bit seven in each 'payload frame' cleared to tell them from the packet header byte (not shown here)                      .
//                                                                                                                                 .
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


#if( SBP_SUPPORT_PACKET_TX ) // support TRANSMISSION of packets (and their assembly in RAM) ?

//---------------------------------------------------------------------------
void SBP_WriteBitgroup( // Appends a 1- to 32-bit value to the packet buffer, "densely packed".
   T_SBP_Instance *pInst, // [out]    pInst->pbDst (always points to the "first payload byte")
                      // [in,out] pInst->iWriteBitIndex (INCREMENTED here)
   uint32_t u32Value, // [in] signed or unsigned integer with anything from 1 to 32 bits .
                      //      Will occupy up to ceil(32/7) = FIVE bytes in pbDst[],
                      //      but the remaining "unused" bits in the last byte
                      //      will be filled in the NEXT call of this function .
   int iBitgroupSize) // [in] Number of bits REALLY USED in u32Value. Range: 1..32 [bits].
   // Notes:
   //   * There's no need for a 'signed integer' variant. Just cast whatever
   //     you need to append to the packet into a 'uint32_t' (aka DWORD),
   //     if an overly pedantic C compiler complains about incomatible data types.
   //   * When compiled with SBP_BUILT_IN_UNIT_TESTS == 1,
   //     there's a built-in "unit test" for SBP_WriteBitgroup()
   //             and it's inverse function, SBP_ReadBitgroup().
   //     The TEST for both functions is implemented in SBP_InitInstance().
   //   * Getting this "bit-fiddling stuff" working properly AND efficient
   //     on a tiny 8-bit microcontroller turned into a nightmare,
   //     thus is was removed from this 'reference implementation' again.
   //
{
   int iBitIndex  = pInst->iWriteBitIndex; // "bit index" into a packet buffer", range = 0 .. 8*SBP_PACKET_BUFFER_SIZE - 1 = 8*32-1 = 255 .
   int iBitOfByte;
   int iByteIndex;  // the "byte index" (into pbDst[]) is the "bit index" / 8 (or bitwise shifted right 3 times)
   int nBitsInByte;
   BYTE b;

   while( ((iByteIndex=(iBitIndex>>3))<SBP_PACKET_BUFFER_SIZE) && (iBitgroupSize>0) ) // more bits to "emit" ?
    {
      // How many "new bits" can be moved into pbDst[iByteIndex] before that byte is 'full' ?
      iBitOfByte = iBitIndex & 7;   // -> 0..7, where 7 would be the SB_PACKET_HEADER_FLAG ("taboo" here)
      nBitsInByte = 7 - iBitOfByte; // -> number of bits that can be moved into pbDst[iByteIndex] in this loop
      // Example: Packet Byte[4], first four bits of Channel B (see "ASCII art" shown further above)
      //         iBitIndex  = 35 = 0x23 (go figure this out .. note: The "packet header byte" is in Packet Byte[0], not shown in the ASCII ART )
      //         iByteIndex = 4  (this is the zero-based BYTE index, into array pbDst[] )
      //         iBitOfByte = 3, which means pbDst[iByteIndex] already contains the example's "A21..A23" (Audio Channel A) in bits 0..2
      //    ->   nBitsInByte = 7 - (35 & 7) = 7 - 3 = 4 .
      // If iBitOfByte was 7, nBitsInByte would be ZERO at this point, which means "begin a new byte" (in the "if" block);
      //    otherwise (iBitOfByte != 7), pbDst[iByteIndex] already "contains a few bits", and bitwise OR is required (in the "else" block).
      if( iBitOfByte==0 ) // beginning a 'new byte', no bit-wise OR required now, may emit up to SEVEN bits in this loop:
       { b = (BYTE)u32Value;
         nBitsInByte = (iBitgroupSize >= 7) ? 7 : iBitgroupSize; // -> 1..7 bits to emit in THIS loop (with iBitgroupSize==0, wouldn't have gotten here)
         b &= (0xFF >> (8-nBitsInByte) );  // for example, with nBitsInByte=7, bitwise-AND with 0x7F (thus avoids setting bit 7, the SB_PACKET_HEADER_FLAG)
       }
      else  // nBitsInByte != 0 -> not a 'new byte', so read back some already-set-bits from the packet buffer:
       { b = pInst->pbDst[iByteIndex]; // <- already contains <iBitOfByte> bits that MUST NOT BE MODIFIED below.
         if( nBitsInByte > iBitgroupSize )
          {  nBitsInByte = iBitgroupSize;   // cannot emit more than what's left in u32Value !
          }
         // The bits already in 'b' must not be shifted but kept 'in place'.
         // Only <nBitsInByte> from u32Value must be shifted LEFT,
         //                    before bitwise ORing them into the 8-bit 'b'.
         // Even though the following looks horrible, only eight bit operations
         //  are required here (only the lower bits from the 32-bit 'uValue'
         //  are involved, and thus immediately cast into a BYTE, to make
         //  C compilers for 8-bit microcontroller like PICs happy) :
         b |= ( ( (BYTE)u32Value & (0xFF >> (8-nBitsInByte) ) ) << iBitOfByte );
         //       |_____________|  |________________________|   |_____________|
         //        next 1..7 (?)    isolate 7..1                 e.g. bitwise shift
         //        "new" bits       "new bits",                  LEFT 3 times,
         //      from the source    e.g bitwise AND              because iBitOfByte=3 [bits]
         //                         with 0x0F if only            were alreay in 'b'
         //                         4 bits are emitted
         //                         in this loop.
         //   |________________________________________________________________|
         //     After this bit-fiddling ("new" and "old" bits in byte 'b'),
         //     the maximum value in b cannot exceed 0x7F = 0b01111111 .
       }
      pInst->pbDst[iByteIndex] = b;  // write the 'new' byte, or write back the 'modified' byte (with a few 'new bits' ORed)
      iBitgroupSize -= nBitsInByte;  // decrement the REMAINING number of bits by 1..7 (not 8)
      u32Value >>= nBitsInByte;      // shift the next bits in place, for the next loop (THIS is a 32-bit operation)
      iBitIndex += nBitsInByte;      // increment the BIT INDEX into the packet buffer (but this was only the 7-bit PAYLOAD) ...
      // If iBitIndex just reached bit SEVEN in a byte of the payload,
      // increment it by one more step (bit) to skip that bit:
      if( (iBitIndex & 7) == 7 )
       { ++iBitIndex; // skip bit SEVEN ( avoid SB_PACKET_HEADER_FLAG occurring in the payload)
       }
    } // end while < more bits or bitgroups to "emit" into the packet buffer >

  pInst->iWriteBitIndex = iBitIndex; // write the incremented BIT-INDEX back

} // end SBP_WriteBitgroup()

#endif // SBP_SUPPORT_PACKET_TX ?



#if( SBP_SUPPORT_PACKET_RX ) // support RECEPTION of packets (and parsing them) ?

//---------------------------------------------------------------------------
uint32_t SBP_ReadBitgroup( // Reads the next 1- to 32-bit integer from a "densely packed" SEVEN-BIT payload.
   T_SBP_Instance *pInst, // [in]     pInst->pbSrc (always points to the "first payload byte")
                          // [in]     pInst->nReadBitsAvailable,
                          // [in,out] pInst->iReadBitIndex (read and INCREMENTED here).
   int iBitgroupSize )    // [in] what the name indicates.. number of "payload"-bits
                          // TO READ from the packet buffer, into the returned value,
                          // ranging from 1 to 32 [bits, not bytes] .
   // [return] : u32Value = integer ("bitgroup" value) read from the packet buffer
   //                       with anything from 1 to 32 bits .
   //                       If the caller wants the result as a SIGNED integer,
   //                       use SBP_ReadBitgroup_Signed() instead.
   //
   // Note: When compiled with SBP_BUILT_IN_UNIT_TESTS == 1,
   //       there's a built-in "unit test" for SBP_ReadBitgroup(),
   //       and it's inverse function, SBP_WriteBitgroup().
   //       The TEST for both functions is implemented in SBP_InitInstance().
{
  int iBitIndex  = pInst->iReadBitIndex; // absolute "bit index into the packet buffer", 0 .. 8*SBP_PACKET_BUFFER_SIZE - 1 = 8*32-1 = 255 .
  int iBitOfByte;
  int iByteIndex;  // the "byte index" (into pbSrc[], treated like an array) is the "bit index" / 8 (or bitwise shifted right 3 times)
  int nBitsInByte, nBitsInResult = 0;
  BYTE b;
  uint32_t u32Value = 0; // 32-bit result

  while( (iBitIndex<pInst->nReadBitsAvailable) && (iBitgroupSize>0) ) // more bits available, and more bits to "read into u32Value" ?
   {
     iByteIndex = iBitIndex >> 3;
     iBitOfByte = iBitIndex & 7;  // -> 0..7, where 7 would be the SB_PACKET_HEADER_FLAG ("not a payload bit").
     // If iBitOfByte was 0, the next bit (to read into u32Value)
     //    would already be in bit ZERO of pbSrc[iByteIndex].
     //    But due to the 'dense packing', that's often NOT the case.
     b = pInst->pbSrc[iByteIndex]; // get the next ONE TO SEVEN bits from a packet buffer

     // How many bits can be extracted from 'b' before that byte 'runs empty' for our bitgroup ?
     nBitsInByte = 7 - iBitOfByte;
     if( nBitsInByte > iBitgroupSize ) // more bits available in 'b' than still missing in u32Value ?
      {  nBitsInByte = iBitgroupSize; // limit to the remaining number of bits to read into u32Value
      }
     b >>= iBitOfByte; // right-align the few 'wanted' bits (an 8 bit operation).
           // The lowest-valued 'wanted' bit is in bit ZERO  of 'b' now,
           //  but the HIGHER-valued bits in 'b' may belong to another bitgroup.
           //  Thus "mask them out" before combining 'b' with u32Value:
     b &= (0xFF>>(8-nBitsInByte)); // mask out bits in 'b' that don't belong to the wanted bitgroup
           // Combine those 'new bits' (1..7, in byte 'b') with u32Value.
           // The loop has begun with the LOWEST part of the bitgroup,
           // so first convert 'b' from 8 to 32 bits, then bitwise
           // shift it LEFT using a 32-bit shift operation:
     u32Value |= (uint32_t)b << nBitsInResult; // <- the only "costly" 32-bit operation for poor old 8-bit microcontrollers :)
     nBitsInResult += nBitsInByte; // If more bits follow in the next loop, THIS is the next number of bitwise-LEFT-shifts
     iBitgroupSize -= nBitsInByte;
     iBitIndex     += nBitsInByte; // increment the BIT INDEX into the entire packet buffer (but this was only the 7-bit PAYLOAD) ...
     // If iBitIndex just reached bit SEVEN in a byte of the 7-bit-payload,
     // increment iBitIndex by one more step (= one bit) to skip "bit seven":
     if( (iBitIndex & 7) == 7 )
      { ++iBitIndex;  // (that's why we need different bit counters for the BITGROUP (u32Value) and the PACKET BUFFER)
      }

   } // end while < more bits or bitgroups to "read" from the packet buffer >

  // Put the INCREMENTED packet-buffer-bit-index back where it belongs,
  //     so the NEXT call of SBP_ReadBitgroup() knows
  //     from where to read the next bitgroup (in the seven-bit payload) :
  pInst->iReadBitIndex = iBitIndex;


  return u32Value;
} // end SBP_ReadBitgroup()

//---------------------------------------------------------------------------
long SBP_ReadBitgroup_Signed( // Almost the same, but SIGNED
   T_SBP_Instance *pInst, // [in,out] : see SBP_ReadBitgroup()
   int iBitgroupSize )    // [in] number of "payload"-bits to read,
                          // ranging from 1 to 32 [bits, not bytes] .
{ long i32Result = (long)SBP_ReadBitgroup( pInst, iBitgroupSize );
  // Expand the SIGN-bit into the long integer (in all its 32-bit glory) :
  if( i32Result & (1L<<(iBitgroupSize-1) ) ) // MSBit *of the bitgroup* set -> value is negative !
   {  i32Result |= (0xFFFFFFFFL << iBitgroupSize );
      // Example: With iBitgroupSize = 24, the SIGN BIT is in bit 23 (zero-based
      //          bit number) of i32Result .
      //          The sign-extenstion-mask, bitwise ORed to make the 32-bit
      //          result negative (above), would be 0xFF000000 in this case .
   }
  return i32Result;
} // end SBP_ReadBitgroup_Signed()

#endif // SBP_SUPPORT_PACKET_RX ?




#if( SBP_SUPPORT_PACKET_TX ) // support TRANSMISSION of packets (and assembling them in RAM) ?

//---------------------------------------------------------------------------
void SBP_BeginWritingPacket( // .. with HEADER BYTES, but not PAYLOAD yet ..
   T_SBP_Instance *pInst, // [in,out] "Seven-Bit Packetizer" instance
   int iContentType )     // [in] SB_PACKET_CONTENT_TYPE_AUDIO_SAMPLE_FORMAT /
                          //      .._TIME_OF_DAY,  .._UNIX_DATE, _GPS_NMEA,
                          //      .._AUDIO, .._ASCII, and whatever may come...
   // Called from any function to assemble a SHORT packet in the instance's
   // packet buffer, i.e. pInst->bPacketBuffer[], using as many calls of
   // SBP_WriteBitgroup(). When done (payload complete), call
   // SBP_EndWritingPacket() to update the packet's payload length,
   // CRC, or whatever will be added in future.
{ int header_byte_index = 0;
  // Already fill out the header (even if the PAYLOAD LENGTH is still unknown):
  switch( iContentType ) // packet content type indicated in the PACKET HEADER BYTE,
   { // or in an extra / optional byte ? Also, for certain CONTENT TYPES (e.g. "ASCII"),
     // already reserve TWO extra header bytes for a 'longer payload' :
     case SB_PACKET_CONTENT_TYPE_AUDIO : // definitely a SHORT packet,
          // and CONTENT TYPE only needs a two-bit-field in the PACKET HEADER BYTE.
          pInst->bPacketBuffer[ header_byte_index++ ] = SB_PACKET_HEADER_FLAG | SB_PACKET_HDR_TYPE_AUDIO | SB_PACKET_HDR_LENGTH_SHORT_DUMMY;
          break;
     case SB_PACKET_CONTENT_TYPE_ASCII : // will NOT be a SHORT packet,
          // but with the CONTENT TYPE indicated in a two-bit-field of the PACKET HEADER BYTE.
          pInst->bPacketBuffer[ header_byte_index++ ] = SB_PACKET_HEADER_FLAG | SB_PACKET_HDR_TYPE_AUDIO | SB_PACKET_HDR_LENGTH_LONG_PAYLOAD;
          pInst->bPacketBuffer[ header_byte_index++ ] = 0; // placeholder for PACKET LENGTH, lower 7 bits
          pInst->bPacketBuffer[ header_byte_index++ ] = 0; // placeholder for PACKET LENGTH, upper 7 bits
          break;
     case SB_PACKET_CONTENT_TYPE_AUDIO_SAMPLE_FORMAT : // definitely SHORT packets,
     case SB_PACKET_CONTENT_TYPE_TIME_OF_DAY: // with "densely packed payload",
     case SB_PACKET_CONTENT_TYPE_UNIX_DATE  : // thus no need for extra bytes for the LENGTH,
          // but with the CONTENT TYPE indicated in an extra HEADER byte .
          pInst->bPacketBuffer[ header_byte_index++ ] = SB_PACKET_HEADER_FLAG | SB_PACKET_HDR_TYPE_OTHER | SB_PACKET_HDR_LENGTH_SHORT_DUMMY;
          pInst->bPacketBuffer[ header_byte_index++ ] = (BYTE)iContentType; // packet content type indicated in an extra header byte
          break;
     case SB_PACKET_CONTENT_TYPE_GPS_NMEA:
     case SB_PACKET_CONTENT_TYPE_OTHER   : // assume these packets will be LONG,
     default:  // AND they need the CONTENT TYPE indicated in an extra HEADER byte.
          pInst->bPacketBuffer[ header_byte_index++ ] = SB_PACKET_HEADER_FLAG | SB_PACKET_HDR_TYPE_OTHER | SB_PACKET_HDR_LENGTH_LONG_PAYLOAD;
          pInst->bPacketBuffer[ header_byte_index++ ] = 0; // placeholder for PACKET LENGTH, lower 7 bits
          pInst->bPacketBuffer[ header_byte_index++ ] = 0; // placeholder for PACKET LENGTH, upper 7 bits
          pInst->bPacketBuffer[ header_byte_index++ ] = (BYTE)iContentType; // packet content type indicated in an extra header byte
          break;
   } // end switch( iContentType )
  // Now (after the above switch/case statement), we know the number of HEADER BYTES,
  // thus we know where the (usually 'densely packed') payload begins.
  // So prepare appending bitgroups of any size (from 1 to 32 bits per bitgroup),
  // using e.g. SBP_WriteBitgroup(), or maybe (in a microcontroller firmware)
  // something 'highly speed-optimized', e.g. for sending audio samples.
  pInst->pbDst = pInst->bPacketBuffer;
  pInst->iWriteBitIndex = 8 * header_byte_index; // "BIT-INDEX" (!) into the payload, appended by the caller
  pInst->nWriteBitsMax  = SBP_PACKET_BUFFER_SIZE * 8/* bits ! */;
  // Next steps to "assemble" a transmitted packet in memory :
  //  * Call SBP_WriteBitgroup() repeatedly, for any payload to append;
  //  * Call SBP_EndWritingPacket() [this will update the PACKET LENGTH in the header, and maybe more]
  //  * Call your UART-block-transmit-function, to send pInst->bPacketBuffer,
  //         with the number of BYTES = (iWriteBitIndex+7) / 8
  //      ,-----------------------------|_____________________|
  //      '--> packet size measured in BITS, "rounded up", then converted to size in BYTES.
  //    For convenience, SBP_EndWritingPacket() returns the TOTAL number of
  //    bytes to send (via UART), including the packet header bytes, payload,
  //    and (maybe, in a future version) an optional CRC (Cyclic Redundancy Code).
} // end SBP_BeginWritingPacket()

//---------------------------------------------------------------------------
int SBP_EndWritingPacket( // .. see "Steps to assemble a transmitted packet" in SBP_BeginWritingPacket() !
   T_SBP_Instance *pInst ) // [in,out] "Seven-Bit Packetizer" instance
   // Return value: The TOTAL number of bytes to send,
   //               beginning with the FIRST BYTE in pInst->bPacketBuffer[0],
   //               and ending with the LAST BYTE in pInst->bPacketBuffer[iPacketSizeInBytes-1] .
{ int iPacketSizeInBytes = ( pInst->iWriteBitIndex + 7 ) / 8;

  // iPacketSizeInBytes (calculated above) is the ENTIRE packet, including the
  // 1..4 (?) header bytes. But the SHORT_PAYLOAD_LENGTH or LONG_PAYLOAD_LENGTH
  // (contained as a 5- or 14-bit field in the header itself) only counts
  // the number of PAYLOAD BYTES within the packet, not the total size. Thus:
  int iPayloadSizeInBytes = iPacketSizeInBytes - SBP_GetHeaderSizeInBytes(pInst->bPacketBuffer[0]);

  // If the 5-bit SHORT_PAYLOAD_LENGTH field contains SB_PACKET_HDR_LENGTH_LONG_PAYLOAD,
  // patch iPayloadSizeInBytes into the two bytes immediately after the Packet Header Byte.
  // Otherwise, patch iPayloadSizeInBytes into the 5-bit SHORT_PAYLOAD_LENGTH field itself:
  if( (pInst->bPacketBuffer[0] & SB_PACKET_HDR_LENGTH_MASK) == SB_PACKET_HDR_LENGTH_LONG_PAYLOAD )
   { // SBP_BeginWritingPacket() has decided to use a LONG_PAYLOAD_LENGTH :
     pInst->bPacketBuffer[1] = iPayloadSizeInBytes & 0x7F; // lower SEVEN(!) bits first..
     pInst->bPacketBuffer[2] = (iPayloadSizeInBytes >> 7) & 0x7F; // upper SEVEN(!) bits last
   }
  else // SBP_BeginWritingPacket() has decided to use a LONG_PAYLOAD_LENGTH :
   { // If, in this case, iPayloadSizeInBytes is too large for 5 bits,
     // we've made a mistake somewhere (e.g. an AUDIO SAMPLE packet got too large,
     // for the given number of channels and the given number of samples per channel.
     // Try to fix that in SBP_BeginWritingPacket(), because all it needs to know
     // is in the T_SBP_Instance (structure) !
     pInst->bPacketBuffer[0] = (pInst->bPacketBuffer[0] & (~SB_PACKET_HDR_LENGTH_MASK/*0x1F*/) )
                             | ((BYTE)iPayloadSizeInBytes & SB_PACKET_HDR_LENGTH_MASK );
     // ,------------------------------'
     // '--> *payload* size, not *packet* size, to keep those numbers as low as possible.
   }

  // That's all .. the packet in pInst->bPacketBuffer[0 .. iPacketSizeInBytes-1]
  //               is "ready for transmission"  ...
#if( SBP_COMPILING_SPECLAB )
  ++pInst->dwTxPacketCounter; // count the packet as "transmitted" (only for 'statistics')
#endif // SBP_COMPILING_SPECLAB ?

  return iPacketSizeInBytes;
} // end SBP_EndWritingPacket()

//---------------------------------------------------------------------------
int SBP_WritePacket_AudioSampleFormat( // ... into the PACKET BUFFER, for transmission
   T_SBP_Instance *pInst ) // [in,out] "Seven-Bit Packetizer" instance :
   // [in] pInst->nBitsPerSingleSample, pInst->nChannelsPerSamplePoint,
   //      pInst->iSampleDataType,      pInst->dblSampleRate_Hz, .. (?)
   // [out] pInst->bPacketBuffer[]
   // [return] number of bytes to send (entire packet)
{
  long integer_part, fractional_part;

  SBP_BeginWritingPacket( pInst, SB_PACKET_CONTENT_TYPE_AUDIO_SAMPLE_FORMAT );
  // '--> pInst->iWriteBitIndex = 16 (ok; PACKET_HEADER_BYTE byte plus
  //                                  CONTENT_TYPE in an EXTRA byte of the header)
  SBP_WriteBitgroup( pInst, pInst->nBitsPerSingleSample,    7/* payload bits*/ );
  SBP_WriteBitgroup( pInst, pInst->nChannelsPerSamplePoint, 7/* payload bits*/ );
  SBP_WriteBitgroup( pInst, pInst->iSampleDataType,         7/* payload bits*/ );
  // '--> pInst->iWriteBitIndex = 16 + 3*8 = 40 bits, ok; includes the skipped 'bits SEVEN')
  SBP_DoubleToFixedPoint21( pInst->dblSampleRate_Hz, &integer_part, &fractional_part );
  SBP_WriteBitgroup( pInst, integer_part, 21/*payload bits*/ );
  // '--> pInst->iWriteBitIndex = 48 + 3*8 = 64 bits, ok... )
  // > "'fractional part', if for some reason the sender cannot achive an integer number of samples".
  // We'll send the fractional part with 21 binary places, 1 / 2^21 = 4.7E-7,
  // thus "ignore" anything smaller than that :
  if( fractional_part != 0 ) // sacrifice THREE FRAMES (in the serial port) for the fractional part ?
   { SBP_WriteBitgroup( pInst, fractional_part, 21/*payload bits*/ );
     //           '--> pInst->iWriteBitIndex = 64 + 3*8 = 88 bits, ok... )
     // This has "blown up" the total packet size from 8 to 11 bytes.
     // This is only possible because this "optional field" (3 bytes in the
     // payload for the 21-bit-group) is the LAST FIELD in the packet's payload.
     // Otherwise, the receiver / packet decoder can't tell from the PACKET LENGTH
     // which fields are really present !
   }
  return SBP_EndWritingPacket( pInst );  // -> returned iPacketSizeInBytes = 8 or 11; ok.
} // end SBP_WritePacket_AudioSampleFormat()



//---------------------------------------------------------------------------
int SBP_WritePacket_AudioSample( // ... into the PACKET BUFFER, for transmission
   T_SBP_Instance *pInst ) // [in,out] "Seven-Bit Packetizer" instance :
   // [in] pInst->fltSampleBuffer[ SBP_MAX_AUDIO_CHANNELS ]
   //   or pInst->i32SampleBuffer[ SBP_MAX_AUDIO_CHANNELS ],
   //      pInst->nBitsPerSingleSample,   -, "variable" sample format
   //      pInst->nChannelsPerSamplePoint, | as previously announced via
   //      pInst->iSampleDataType,        -' SBP_WritePacket_AudioSampleFormat()
   // [out] pInst->bPacketBuffer[]
   // [return] number of bytes to send (entire packet, including header)
{
  int iChannel, nChannels;
  long i32SampleValue;
#if(SBP_USE_FLOATING_POINT)
  float flt;
#endif

  nChannels = pInst->nChannelsPerSamplePoint;
  if( nChannels > SBP_MAX_AUDIO_CHANNELS ) // oops.. misconfigured SevenBitPackets.h !
   {  nChannels = SBP_MAX_AUDIO_CHANNELS;  // limit this to our sample-buffer-capacity
   }

  SBP_BeginWritingPacket( pInst, SB_PACKET_CONTENT_TYPE_AUDIO );
   // '--> pInst->iWriteBitIndex = 8 (ok; only the PACKET_HEADER_BYTE, no extra header bytes)
  for(iChannel=0; iChannel<nChannels; ++iChannel )
   {
#   if(SBP_USE_FLOATING_POINT) // Samples internally stored in floating point format ?
     flt = pInst->fltSampleBuffer[iChannel]; // [in] samples normalized to +/- 1.0, floating point, as used in Spectrum Lab
     if( flt >  1.0 )   flt = 1.0;     // "saturate" ...
     if( flt < -1.0 )   flt = -1.0;    //  .. to prevent integer overflow below
     if( (pInst->iSampleDataType == SBP_DATA_TYPE_UNSIGNED_LE )
       ||(pInst->iSampleDataType == SBP_DATA_TYPE_UNSIGNED_BE ) )
      { // offset 'flt' to make it UNSIGNED:
        flt += 1.0;  // -> flt now ranges from 0.0 to 2.0,
        //    thus the scaling factors below are "almost" ok
        //    (for 8-bit unsigned, the max will be 2*127 = 254 .. oh well.)
      }
     switch( pInst->nBitsPerSingleSample )
      { case  8: flt *= 127.0;      break; // scale for 8-bit signed integer
        case 10: flt *= 511.0;      break; // scale for 10-bit signed integer
        case 12: flt *= 2047.0;     break; // scale for 12-bit signed integer
        case 14: flt *= 8191.0;     break; // scale for 14-bit signed integer
        case 16: flt *= 32767.0;    break; // scale for 14-bit signed integer
        case 21: flt *= 1048576.0;  break; // scale for 21-bit signed integer
        case 24: flt *= 8388607.0;  break; // scale for 24-bit signed integer
        default: flt *= (float)(0xFFFFFFFFUL >> (uint32_t)(32-pInst->nBitsPerSingleSample) ); // spend "more CPU time" for all the rest
                 break;
      } // end switch( pInst->nBitsPerSingleSample )
     i32SampleValue = (long)flt;
#   else  // ! SBP_USE_FLOATING_POINT -> no need to scale from +/- 1.0 to the "number of bits per sample", because
     i32SampleValue = pInst->i32SampleBuffer[iChannel]; // this buffer already holds INTEGER values (with up to 32 bits per sample)
     // Not scaled in this case, because nBitsPerSingleSample will match
     // whatever is delivered by the A/D converter in "RIGHT-ALIGNED" format,
     // e.g. -32768 .. +32767 for a hypothetic ADC with bipolar (or differential)
     // input, and 16 bit of resolution.
#   endif // ! SBP_USE_FLOATING_POINT ->
     SBP_WriteBitgroup( pInst, i32SampleValue, pInst->nBitsPerSingleSample/*payload bits*/ );
    } // end for < iChannel .. >
   return SBP_EndWritingPacket( pInst );  // like other 'packetizer' functions,
      // returns the total packet size in BYTES. For 2 * 24 bits "payload",
      // the total packet size was EIGHT .. ok (including ONE header-byte).

} // end SBP_WritePacket_AudioSample()

#endif // SBP_SUPPORT_PACKET_TX  ?


#if( SBP_SUPPORT_PACKET_RX ) // support RECEPTION of packets (and parsing them in RAM) ?
//---------------------------------------------------------------------------
void SBP_ProcessPayload_AudioSample( // ... from the PACKET BUFFER, for reception
   T_SBP_Instance *pInst) // [in] : "densely packed 7-bit payload", see SBP_ReadBitgroup()
   // [in] pInst->nBitsPerSingleSample,   ---, "variable" sample format,
   //      pInst->nChannelsPerSamplePoint,   | as previously announced in a packet with
   //      pInst->iSampleDataType,           | SB_PACKET_CONTENT_TYPE_AUDIO_SAMPLE_FORMAT .
   //      pInst->dbl/i32SampleRate_Hz .  ---'
   // [out] pInst->fltSampleBuffer[ SBP_MAX_AUDIO_CHANNELS ]
   //    or pInst->i32SampleBuffer[ SBP_MAX_AUDIO_CHANNELS ],
   // Called from SBP_ProcessReceivedPacket() on reception of valid packet with
   //             PACKET_HDR_TYPE_AUDIO (two-bit-field in the packet header byte).
   // Originally written to RECEIVE SAMPLES from an external A/D converter
   //                    via 'COM port' in Spectrum Lab.
   // Something simular (or just THIS FUNCTION) will be necessary in a
   //                    microcontroller firmware driving a D/A converter.
{
  int iChannel, nChannels;
  long i32SampleValue;
#if(SBP_USE_FLOATING_POINT)
  float flt;
#endif

  nChannels = pInst->nChannelsPerSamplePoint;
  if( nChannels > SBP_MAX_AUDIO_CHANNELS ) // oops.. misconfigured SevenBitPackets.h !
   {  nChannels = SBP_MAX_AUDIO_CHANNELS;  // limit this to our sample-buffer-capacity
   }

  for(iChannel=0; iChannel<nChannels; ++iChannel )
   {
     i32SampleValue = SBP_ReadBitgroup_Signed( pInst, pInst->nBitsPerSingleSample );
#   if(SBP_USE_FLOATING_POINT) // Samples internally stored in floating point format ?
     // Using pInst->fltSampleBuffer[iChannel], with samples normalized to +/- 1.0 :
     flt = (float)i32SampleValue;  // convert from integer to floating point and normalize:
     switch( pInst->nBitsPerSingleSample ) // assume iSampleDataType is SBP_DATA_TYPE_SIGNED_LE ...
      { case  8: flt *= ( 1.0 / 127.0);      break; // normalize to ~~+/-1 from 8-bit signed int (note the "missing code" -128, assume the range is +/-127)
        case 10: flt *= ( 1.0 / 511.0);      break; // normalize to ~~+/-1 from 10-bit signed int
        case 12: flt *= ( 1.0 / 2047.0);     break; // normalize to ~~+/-1 from 12-bit signed int
        case 14: flt *= ( 1.0 / 8191.0);     break; // normalize to ~~+/-1 from 14-bit signed int
        case 16: flt *= ( 1.0 / 32767.0);    break; // normalize to ~~+/-1 from 16-bit signed int
        case 21: flt *= ( 1.0 / 1048576.0);  break; // normalize to ~~+/-1 from 21-bit signed int
        case 24: flt *= ( 1.0 / 8388607.0);  break; // normalize to ~~+/-1 from 24-bit signed int
        default: flt *= ( 1.0 / (float)(0xFFFFFFFFUL >> (uint32_t)(32-pInst->nBitsPerSingleSample) ) ); // spend "more CPU time" for all the rest
                 break;
      } // end switch( pInst->nBitsPerSingleSample )
     pInst->fltSampleBuffer[iChannel] = flt;
#   else  // ! SBP_USE_FLOATING_POINT -> no need to scale, just store as sign-expanded 32-bit integer
     pInst->i32SampleBuffer[iChannel] = i32SampleValue;
#   endif // ! SBP_USE_FLOATING_POINT ->
    } // end for < iChannel .. >

   // Let the application consume another new multi-channel "sample point",
   //     in pInst->fltSampleBuffer[] or pInst->fltSampleBuffer[].
   // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   //   A microcontroller-based driver for a D/A converter will queue up the
   //   samples in a circular FIFO, and let some timer interrupt send the samples
   //   to the D/A converter, carefully timed to a microsecond. Because when
   //   sending data over a serial port from Windows, blocks of bytes will
   //   arrive 'in bursts' with an unpredictable timing, thus don't immediately
   //   write the samples to the D/A converter in the function called below.
   //   How to 'synchronize' the emission of samples to the sampling rate indicated
   //   by pInst->dblSampleRate_Hz (or pInst->i32SampleRate_Hz for uCs w/o floating point),
   //   and if / how to manage the problem of "slightly different sampling rates"
   //   in the sender and the receiver is entirely up to you.
   // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# if( SBP_USE_CALLBACK_FOR_RECEIVED_AUDIO ) // Invoke 'application callback' via FUNCTION POINTER ?
   if( pInst->pfRxAudioCallback != NULL )
    {  pInst->pfRxAudioCallback( pInst->pvCallbackData, pInst );
       //  ,------------------------------'              '----------,
       // pointer to "anything" passed on to the above callback     |
       //             (in Spectrum Lab: a CSound object)            |
       //    ,------------------------------------------------------'
       //    '-> "instance data" with all the callee needs to know
       //        to process the sample point, including number of
       //        channels, sampling rate, and who-knows-what (timestamp?) .
    }
# else // ! SBP_USE_CALLBACK_FOR_RECEIVED_AUDIO : no 'callback' via function pointer,
       //   but just a plain boring normal subroutine call, to some unknown external
       //   module. Intended for stoneage little 8-bit "PIC" microcontrollers
       //   where real 'function pointers' are *impossible* (hard to believe but true).
   SBP_ProcessReceiveAudio( pInst ); // <- subroutine NOT IMPLEMENTED in SevenBitPackets.c !
# endif // SBP_USE_CALLBACK_FOR_RECEIVED_AUDIO ?


} // end SBP_ProcessPayload_AudioSample()

#endif // SBP_SUPPORT_PACKET_RX  ?



//---------------------------------------------------------------------------
// Optional 'utility functions' (not so interesting, thus near the end..)
//---------------------------------------------------------------------------


// Conversion of floating point ("double", as internally used for the SAMPLING RATE,
//   for SB_PACKET_CONTENT_TYPE_TIME_OF_DAY, etc into FIXED POINT with a 21(!) bit fractional part:
#if( SBP_USE_FLOATING_POINT ) // May the implementation use 'float' and 'double' at all ?
//---------------------------------------------------------------------------
void SBP_DoubleToFixedPoint21( // "twentyone bits of resolution for the FRACTIONAL part":
         double d, // [in] 64-bit "double", here: ranging ~~ +/-(2^20-1) = 1048575,
                   //      if the resulting integer part shall fit inside 21 bits.
                   //      Fractional part circa +/-0.999999 (good for SIX decimal places).
         long *pi32IntegerPart,    // [out,optional] integer part (signed)
         long *pi32FractionalPart) // [out,optional] signed fractional part,
                                   //   scaled for 21 bits for easy packing.
                                   // (scaling factor exactly 2^20 = 1048576
                                   //  thus value range -1048576 .. +1048575 )
{
  // Naively multiplying the 'double' by 2^21, then truncating the result
  //         to a 32-bit integer won't work - the MSBits would "fall out".
  //  64 bit integers are not available in some of the intended targets
  //         (strangely, some embedded C libaries supported 'double' but no
  //          'int64' in any flavour).
  // Thus, carefully 'truncate' to integer, subtract to get the 'rest',
  // then scale the remaining fractional part (0 ... 0.999999x) for 20(!) bits.
  // We don't even need 'math.h' for this, because (about "C"):
  // > Casting to an int will truncate toward zero.    [WB: that's what we want]
  // > floor() will truncate toward negative infinite. [WB: that's NOT what we want]
  //
  long int_part   = (long)d;
  long fract_part = (long)( 1048576.0/* 2^20 */ * (d - (double)int_part) );
       // Note: If 'd' was negative and had a nonzero fractional part,
       //       then 'fract_part' will be negative, too ! That's a difference
       //       to the classic FIXED POINT format, where there is only
       //       ONE common sign-bit for the entire number (e.g. 'Q16.16' as
       //       used ARM, with 1 bit SIGN, 15 bit integer part,
       //       and 16 bit fractional part. OUR 'fract_part' may be NEGATIVE !
       //       This "FixedPoint21" format was designed this way so the CALLER
       //       may decide to simply LEAVE THE FRACTIONAL PART AWAY in the packet,
       //       if its value is low, without having to modify the INTEGER part.
       // Example (you will find this in the UNIT TEST, when compiled for Spectrum Lab):
       //       d = -1048575.000001 ["samples per second"]
       //         -> int_part   = -1048575
       //            fract_part = (long)( 1048576 * (-1048575.000001 - (-1048575) ) )
       //                       = (long)( -1.04858 ) = -1
       //       (again: casting double to some integer will truncate TOWARD ZERO)
       //   -> Caller decides that sending such a small fractional part
       //      isn't worth wasting packet space (and serial port's bandwidth),
       //      and leaves the three last optional bytes, with the FRACTIONAL PART,
       //      in the short packet away (e.g. in SB_PACKET_CONTENT_TYPE_TIME_OF_DAY,
       //      or SB_PACKET_CONTENT_TYPE_AUDIO_SAMPLE_FORMAT).
  if( pi32IntegerPart != NULL )        // caller wants the INTEGER part ?
   { *pi32IntegerPart = int_part;      // .. pass it back (not limited here yet)
   }
  if( pi32FractionalPart != NULL)      // caller wants the FRACTIONAL part ?
   { *pi32FractionalPart = fract_part; // .. here it is (fits in TWENTYONE bits)
   }

} // end SBP_DoubleToFixedPoint21()

//---------------------------------------------------------------------------
double SBP_FixedPoint21ToDouble( // inverse to SBP_DoubleToFixedPoint21() .. details there.
        long integer_part,       // [in] signed integer part
        long fractional_part )   // [in] signed(!) fractional part, with scaling factor 2^20
{ double d = (double)integer_part;
  if( fractional_part != 0 )  // don't waste time if there's nothing to scale and add:
   { d += (double)fractional_part * ( 1.0 / 1048576.0/* 2^20 */ );
     // (MULTIPLY may be faster than DIVIDE, if the C compiler isn't smart enough
     //  to recognize the CONSTANT scaling factor. On a PC, this doesn't matter.)
   }
  return d;
} // end SBP_FixedPoint21ToDouble()

#endif // SBP_USE_FLOATING_POINT ?




#if( SBP_SUPPORT_PACKET_RX ) // support RECEPTION of packets (and processing them in RAM) ?

//---------------------------------------------------------------------------
void SBP_ParseReceivedString( T_SBP_Instance *pInst, char *psz7BitASCII )
{
  (void)pInst;            // prevent 'unused argument' warning ..
  (void)psz7BitASCII;     // .. as long as this function is just a dummy

} // end SBP_ParseReceivedString()


//---------------------------------------------------------------------------
void SBP_ProcessReceivedPacket(
   T_SBP_Instance *pInst,  // [in,out] "Seven-Bit Packetizer" instance
                           //      '--> depending on the packet type,
                           //           subroutines called from here
                           //           may modify the 'audio sample format' (info),
                           //           or pass that info on to the application
                           //           via callback function, etc.
   BYTE *pbPacket,         // [in] pointer to the packet-HEADER, followed by PAYLOAD:
                           //   pbPacket[ 0 .. nHeaderBytes-1 ] = header,
                           //   pbPacket[ nHeaderBytes .. + nHeaderBytes+nPayloadBytes-1 ] = payload,
                           //   with only SEVEN bits in each byte actually used.
                           //   Important: pbPacket points into the caller's
                           //         RECEIVE BUFFER. For good reasons, the received
                           //         'payload' has NOT been copied into
                           //          pInst->bPacketBuffer[] by the caller !
   int  nHeaderBytes,     // [in] size of the packet-HEADER (in bytes)
   int  nPayloadBytes )   // [in] size of the raw packet-PAYLOAD (in bytes),
                          //      already extracted from the header,
                          //      or COUNTED(!) in the payload by the caller.
{
  int   iPacketContentType;
  long  integer_part, fractional_part; // .. for FIXED-POINT with "21 bit fractional part"

  // Assume the received packet has a "densely packet payload", thus for the
  // reference implementation, prepare reading bitgroups of any size (1..32 bits)
  // from the RECEIVED PAYLOAD (skipping bit SEVEN in each byte of the payload):
  pInst->pbSrc = pbPacket + nHeaderBytes; // source for SBP_ReadBitgroup() : here not necessarily the "packet buffer" !
  pInst->iReadBitIndex      = 0; // "BIT-INDEX" (!) into the payload, used by SBP_ReadBitgroup() or SBP_ReadBitgroup_Signed()
  pInst->nReadBitsAvailable = nPayloadBytes * 8; // number of bits "available" for SBP_ReadBitgroup(), etc


  switch( pbPacket[0] & SB_PACKET_HDR_TYPE_MASK ) // branch by the two-bit PACKET_HEADER_TYPE ...
   { case SB_PACKET_HDR_TYPE_AUDIO :  // it's a 'short' audio sample point
        // Here, in the 'reference implementation' for Spectrum Lab,
        // just decode what we've got into a SINGLE SAMPLE POINT in RAM :
        SBP_ProcessPayload_AudioSample( pInst );
        break;
     case SB_PACKET_HDR_TYPE_OTHER :  // the type is specified by an extra byte,
        // that follows after the 14-bit "long packet length",
        // or (if it's a "short packet") immediately after the header byte:
        if( (pbPacket[0] & SB_PACKET_HDR_LENGTH_MASK) == SB_PACKET_HDR_LENGTH_LONG_PAYLOAD )
         { // the PACKET HEADER BYTE indicates a 'short payload', thus the "content type"...
           iPacketContentType = pbPacket[3]; // .. follows after header byte [0] and 14-bit 'payload length' [1,2]
         }
        else // the PACKET HEADER BYTE indicates a 'short payload', thus the "content type"...
         { iPacketContentType = pbPacket[1]; // .. follows mmediately after the header byte [0]
         }
        switch( iPacketContentType )
         { case SB_PACKET_CONTENT_TYPE_AUDIO_SAMPLE_FORMAT:
                // See SpecLab/html/Seven_Bit_Packets_for_the_serial_port.htm#densely_packed_audio_format ...
                if( nPayloadBytes >= 6 )  // the following bitgroups are mandatory:
                 { pInst->nBitsPerSingleSample    = SBP_ReadBitgroup( pInst,  7/*bits*/ );
                   pInst->nChannelsPerSamplePoint = SBP_ReadBitgroup( pInst,  7/*bits*/ );
                   pInst->iSampleDataType         = SBP_ReadBitgroup( pInst,  7/*bits*/ );
                   integer_part = SBP_ReadBitgroup_Signed( pInst, 21/*bits*/ );  // .. of the sampling rate
                 } // end if < got the min payload for a SB_PACKET_CONTENT_TYPE_AUDIO_SAMPLE_FORMAT > ?
                if( nPayloadBytes >= 9 )  // the following bitgroup is optional:
                 { // 21-bit optional FRACTIONAL part, sent if the sampling rate isn't an integer:
                   fractional_part = SBP_ReadBitgroup_Signed( pInst, 21/*bits*/ );
                 }
                else
                 { fractional_part = 0;
                 }
#              if(SBP_USE_FLOATING_POINT)
                pInst->dblSampleRate_Hz = SBP_FixedPoint21ToDouble( integer_part, fractional_part );
#              else  // for small microcontrollers without floating point support (or too little ROM) :
                pInst->i32SampleRate_Hz = integer_part;
#              endif // SBP_USE_FLOATING_POINT ?
                break; // end case SB_PACKET_CONTENT_TYPE_AUDIO_SAMPLE_FORMAT
           case SB_PACKET_CONTENT_TYPE_TIME_OF_DAY :
                break;
           case SB_PACKET_CONTENT_TYPE_UNIX_DATE   :
                break;
           case SB_PACKET_CONTENT_TYPE_GPS_NMEA    :
                break;
           default:
                break;
         } // end switch( iPacketContentType )
        break;
     case SB_PACKET_HDR_TYPE_ASCII :  // the payload is an ASCII string, 7 bits per character
        // There's no extra byte for the PACKET_CONTENT_TYPE, because
        // a KEYWORD at the begin of the line (text) shall indicate
        // what to do with the rest of the line, e.g. with a list of Key=Value pairs.
        // To simpify parsing the string as a zero-byte-terminated 'C' string,
        // copy it into the PACKET BUFFER, where it can be safely 'terminated'.
        // (Don't zero-terminate the string directly in the RX buffer,
        //  because that may violate the caller's unknown array limits) .
        if( pbPacket[ nHeaderBytes+nPayloadBytes-1 ] == 0x00 )
         { // Ok, 7-bit ASCII payload is already a zero-terminated "C"-string
           //      -> easy to parse; should we ever decide to parse it :)
           SBP_ParseReceivedString( pInst, (char*)pbPacket + nHeaderBytes );
         }
        break;
     case SB_PACKET_HDR_TYPE_RESERVED: // the type is 'reserved for future use'
        break;
   } // end switch( pbPacket[0] & SB_PACKET_HDR_TYPE_MASK )

} // end SBP_ProcessReceivedPacket()


//---------------------------------------------------------------------------
int SBP_ProcessReceivedData( // .. from the serial port ..
   T_SBP_Instance *pInst, // [in,out] "Seven-Bit Packetizer" instance
   BYTE *pbSource,        // [in] BYTE * pbSource, source block (not necessarily the BEGIN of a "packet")
   int  nBytesRcvd        // [in] number of BYTES in the above block (not necessarily a complete "packet", but ANY number of BYTES received since the last call)
# if( SBP_USE_CALLBACK_FOR_RECEIVED_AUDIO )
   , T_SBP_RxAudioCallback pfRxAudioCallback, // [in] callback for received audio samples (may be NULL = "no callback" for testing)
     void *pvCallbackData  // [in] 'user data' passed on to the above callback. In SpecLab, the address of a CSound object.
# endif // SBP_USE_CALLBACK_FOR_RECEIVED_AUDIO ?
   )
  // Called from Spectrum Lab after the thread 'chewing on the serial port'
  // has delivered any number of bytes (limited to the number of bytes
  // that Windows can handle .
  // Return value : Number of bytes ACTUALLY CONSUMED from the
  //                input block (pbSource, max nBytesRcv) .
  //                The reason for NOT 'consuming' all bytes
  //                is explained below:
  // Similar as the Ogg/Vorbis decoder implemented in Spectrum Lab,
  // the 'de-packetizer' may not be able to proccess ALL BYTES
  // "offerered for processing" by the caller. This is done on purpose,
  // to avoid having to implement another layer of byte-buffering here.
  // This de-packetizer may reject a few bytes after the end of a packet,
  // hoping that in THE NEXT call, one or more complete packets can be exctacted,
  // with the input in a simple contiguous block (not a CIRCULAR BUFFER).
{
  int   nHeadersFound = 0;
  int   nHeaderBytes, nPayloadBytes, nBytesLeft;
  BYTE *pbSrc          = pbSource;
  BYTE *pbSrcEndstop   = pbSource + nBytesRcvd;
  BYTE *pbPacketHeader = pbSource;
  BYTE *pbNextHeader   = pbSource;
  BYTE bPacketHeaderByte;

  //
  pInst->pfRxAudioCallback = pfRxAudioCallback;
  pInst->pvCallbackData    = pvCallbackData;


  // Look for the next PACKET_HEADER_BYTE .
  // Throw away all bytes before the packet header because,
  // in the PREVIOUS call (*if* there was a previous call),
  // we have only consumed the bytes up to, and including,
  // the LAST byte in the packet. If there was the begin
  // of a not-yet-complete packet, the value returned from
  // SBP_ProcessReceivedData() was LESS than nBytesRcvd, and the caller
  // (Spectrum Lab) has LEFT THOSE FEW REMAINING BYTES in its receive-buffer.
  while( pbSrc < pbSrcEndstop )
   { if( *pbSrc & SB_PACKET_HEADER_FLAG )
      { // Ok, got another PACKET_HEADER_BYTE ..
        ++nHeadersFound;
        pbPacketHeader = pbNextHeader = pbSrc;
        nBytesLeft     = pbSrcEndstop - pbSrc; // number of bytes LEFT to process (in the received data block)
        bPacketHeaderByte = *(pbSrc++);
        nHeaderBytes  = SBP_GetHeaderSizeInBytes( bPacketHeaderByte ); // determine the number of HEADER BYTES (1..3)
        // Got enough bytes to PROCESS this packet ?
        nPayloadBytes = 0; // 'payload length' will be determined below
        switch( bPacketHeaderByte & SB_PACKET_HDR_LENGTH_MASK )
         { case SB_PACKET_HDR_LENGTH_NOT_INDICATED : // length not indicated ->
              { // must SEARCH for the *next* packet-header-byte
                // to determine the total length of THIS packet !
                // (For the SENDER, this makes things easy; for a RECEIVER it doesn't)
                BYTE *pbSrc2 = pbSrc;
                while( pbSrc2 < pbSrcEndstop )
                 { if( *pbSrc2 & SB_PACKET_HEADER_FLAG )
                    { // Ok, already saw the *next* packet's PACKET_HEADER_BYTE,
                      // thus the CURRENT packet (without an "indicated length")
                      // is ready for processing .
                      nPayloadBytes = (pbSrc2 - pbSrc); // add the "payload size" counted in this loop
                      //  ,------------|______________|
                      //  '--> ZERO if the next byte with SB_PACKET_HEADER_FLAG (bit SEVEN)
                      //       follows immediately after the one above (bPacketHeaderByte).
                      //       That's strange but not an error (may be just an empty
                      //       'flow control packet' in future versions..)
                      break;
                    }
                   else
                    { ++pbSrc2;
                      if( pbSrc2 >= pbSrcEndstop )
                       { return pbPacketHeader - pbSource; // look at THIS header in the next call again
                         // (hope we've got enough bytes then..)
                       }
                    }
                 }
              } break; // end case SB_PACKET_HDR_LENGTH_NOT_INDICATED
           case SB_PACKET_HDR_LENGTH_LONG_PAYLOAD : // length in two subsequent bytes
                if( nBytesLeft < 3 ) // haven't got enough bytes yet (for header-byte + two "payload length bytes")
                 { return pbPacketHeader - pbSource; // look at THIS header in the next call again
                 }
                nPayloadBytes  = *(pbSrc++) & 0x7F;         // payload length, lower 7 bits
                nPayloadBytes |= (*(pbSrc++) & 0x7F) << 7;  // payload length, upper 7 bits
                break;
           default :   //  any 'short' packet, with length indicated in the packet-header-byte:
                nPayloadBytes += bPacketHeaderByte & SB_PACKET_HDR_LENGTH_MASK;
                break;
         } // end switch( bPacketHeaderByte & SB_PACKET_HDR_LENGTH_MASK )

        // Now that the TOTAL packet length (not just the PAYLOAD) is known,
        // do we have 'all we need' available, or wait until more bytes
        // have been received ?
        if( nBytesLeft < (nHeaderBytes + nPayloadBytes ) )
         { return pbPacketHeader - pbSource; // look at THIS header in the next call again
         }

        // Arrived here ? Bingo, got enough bytes in the caller's RX-buffer
        // to process the next packet. We already know the number
        // of HEADER BYTES, and the number of PAYLOAD BYTES,
        // so just pass those info to the function below:
        SBP_ProcessReceivedPacket( pInst, pbPacketHeader, nHeaderBytes, nPayloadBytes );

        // Regardless of the packet at pbPacketHeader has been successfully processed or not,
        //  skip it, so the loop continues looking for the NEXT packet at pbSrc :
        pbSrc = pbNextHeader = pbPacketHeader + nHeaderBytes + nPayloadBytes;

#      if( SBP_COMPILING_SPECLAB )
        ++pInst->dwRxPacketCounter; // count the packet as "received" (only for 'statistics')
#      endif // SBP_COMPILING_SPECLAB ?

      } // end if < saw another PACKET_HEADER_BYTE at *pbSrc >
     else // *pbSrc is NOT the expected 'header byte' so skip it
      { ++pbSrc;
#      if( SBP_COMPILING_SPECLAB )
        ++pInst->dwErrorCounter; // count anything "received outside packets" as ERROR (only for 'statistics')
#      endif // SBP_COMPILING_SPECLAB ?
      }
   } // end while( pbSrc < pbSrcEndstop )

  // Arrived here: Escaped from the 'packet searching loop' ...
  if( nHeadersFound <= 0 ) // didn't find ANY header-byte in the loop ->
   { return nBytesRcvd;    // discard the ENTIRE input-block
   }
  // Arrived here ? pbNextHeader points to the byte that we WANT TO
  //    process in the NEXT call (we cannot process
  //    it in THIS call, because not enough bytes received yet
  //    to process an entire packet).
  //    Thus, for the caller, indicate how many bytes we 'consumed' :
  return pbNextHeader - pbSource;  // C pointer acrobatics .. dangerous but fast !

} // end SBP_ProcessReceivedData()

#endif // SBP_SUPPORT_PACKET_RX ?


#if( SBP_BUILT_IN_UNIT_TESTS )
//---------------------------------------------------------------------------
CPROT int SBP_RunUnitTests( void )
  // Performs a rather simplistic 'unit test', especially for the bit-fiddling
  // functions to pack ("write") integers of any size (1..32 bits)
  //              into "serial frames" with 7 bits of usable payload,
  // and to unpack ("read") those integers back from the 'packet buffer' .
  // Returns THE NUMBER OF ERRORS counted during the tests ("ideally zero").
{
  T_SBP_Instance packetizer, *pInst = &packetizer;
  int i, iPass, iBitgroupIndex, iBitgroupSize;
  uint32_t u32Value, u32ValueRead;  // bitgroup value, 0x00000001 ... 0xFFFFFFFF for 1 ... 32 bits
  long integer_part, fractional_part;
  int nErrors = 0;
  #define MAX_BITGROUPS_PER_PACKET_BUFFER_TEST (SBP_PACKET_BUFFER_SIZE/2) // assume 16-bit-groups on average
  uint32_t u32TestBitgroupValues[ MAX_BITGROUPS_PER_PACKET_BUFFER_TEST ];
  int      iTestBitgroupSizes[ MAX_BITGROUPS_PER_PACKET_BUFFER_TEST ];
#if( SBP_USE_FLOATING_POINT )
  double d1, d2;
#endif // SBP_USE_FLOATING_POINT ?

  SBP_InitInstance( pInst );

  randomize(); // Init random generator. Most "stdlib" implementations have this, some don't.

  // Built-in test for the bit-fiddling stuff : Fill the packet buffer with
  // bitfields of varying sizes (from 1 to 32 bits per bitfield),
  // with varying values, then read the whole thing back, extract all bitgroups
  // and compare their values to the previously written ones (thus the local ARRAYS).
  // Do NOT clear the 'written' packet buffer between passes, because
  //    SBP_WriteBitgroup() shall not just "bitwise OR" bits into the buffer,
  //    but REPLACE all bits belonging to a bitgroup (i.e. CLEAR and SET bits).
  for( iPass=0; iPass<20; ++iPass ) // multiple passes with ever-changing TEST VALUES..
   {
     // Prepare writing into "packet buffer" from the start :
     pInst->pbDst = pInst->bPacketBuffer; // let SBP_WriteBitgroup() use the built-in "packet buffer"
     pInst->iWriteBitIndex = 0;
     pInst->nWriteBitsMax  = SBP_PACKET_BUFFER_SIZE * 8/*bits!*/; // MAXIMUM number of bits ( = "endstop") for SBP_WriteBitgroup()

     // Deliberate fill the packet buffer with nonzero bits.
     // SBP_WriteBitgroup() is supposed to CLEAR garbage,
     // except for bit SEVEN in the used bytes of the packet buffer.
     // If it doesn't, we will blame this as an "error" further below.
     for( i=0; i<SBP_PACKET_BUFFER_SIZE; ++i )
      { pInst->bPacketBuffer[i] = 0x7F;
      }
     iBitgroupIndex = 0;
     while( pInst->iWriteBitIndex < ( SBP_PACKET_BUFFER_SIZE*8 - 40/*BITS*/ ) && ( iBitgroupIndex < MAX_BITGROUPS_PER_PACKET_BUFFER_TEST ) )
      { // ,--------------|_______________________________________|
        // '--> for a 32-bit "netto payload", up to FOURTY BITS may be appended
        //      to the 'packet buffer' (with SBP_PACKET_BUFFER_SIZE measured in BYTES).
        //      Thus, for simplicity, use 40 bits less than SBP_PACKET_BUFFER_SIZE.

        // Generate test pattern and WRITE to the packet buffer (as a 7-bit "densely packed" payload):
        if( iPass == 0 ) // In the FIRST pass and it's "verify run",
         { // provide these 'reproducable', constant patterns :
           u32Value      = 0xFFFFFFFF;    // ALL bits in the 1..32 bit-group set, but ..
           iBitgroupSize = 1 + ( iBitgroupIndex & 31 ); // .. variable bitgroup SIZES.
         }
        else if( iPass == 1 ) // In the SECOND pass and it's "verify run",
         { // use another 'reproducable', constant pattern :
           u32Value      = iBitgroupIndex;     // variable values but not random noise, and...
           iBitgroupSize = 1 + ( iBitgroupIndex & 31 );
         }
        else // in all other passes, use pseudo-random "noise" as BITGROUP LENGTHS and BITGROUP DATA:
         { u32Value = _lrand(); // Most "stdlib" implementations have this, some don't. Crap.
           u32Value ^= (u32Value << 17); // make those "upper bits" alive, just in case _lrand() is in fact rand()
                                         // [16 or even just 15 bits, but we want ALL 32 BITS to be "alive"]
           iBitgroupSize = 1 + ( (u32Value >> 5) & 31); // -> "arbitrary length", from 1 to 32 bits
         }
        // Note that, at this point, u32Value may be higher than the number of bits (iBitgroupSize)
        // would permit in an N-bit integer. This is done on purpose, to check
        // if SBP_WriteBitgroup() properly 'masks out' excessive bits,
        // instead of "trashing" neighbouring bits, or even set BIT SEVEN in the output.

        u32TestBitgroupValues[ iBitgroupIndex ] = u32Value;
        iTestBitgroupSizes[ iBitgroupIndex ]    = iBitgroupSize;

        SBP_WriteBitgroup( pInst, u32Value, iBitgroupSize );
        ++iBitgroupIndex;
      } // end while( pInst->iWriteBitIndex < ... )

     // Make sure SBP_WriteBitgroup() hasn't set BIT SEVEN in any byte of the 'packet buffer':
     for( i=0; i<SBP_PACKET_BUFFER_SIZE; ++i )
      { if( pInst->bPacketBuffer[i] & SB_PACKET_HEADER_FLAG )
         { ++nErrors; // oops... SBP_WriteBitgroup() has accidentally set BIT SEVEN !
#         if( 1 && SBP_COMPILING_SPECLAB )// only used during the early development phase:
           DEBUG_EnterErrorHistory( DEBUG_LEVEL_ERROR, 0, UTL_USE_CURRENT_TIME,
              "Error testing SevenBitPackets.c : PACKET_HEADER_FLAGS accidentally set at byte offset %d, valued = 0x%02X",
              (int)i, (int)pInst->bPacketBuffer[i]  );
#         endif // SBP_COMPILING_SPECLAB ?
         }
      }

     // Prepare READING all bitgroups FROM the packet buffer. This emulates
     // processing a "densely packed 7-bit payload" from a received packet.
     pInst->pbSrc = pInst->bPacketBuffer; // let SBP_ReadBitgroup() use the built-in "packet buffer" ..
     pInst->iReadBitIndex = 0; // .. beginning at the FIRST bit in pbSrc[] ..
     pInst->nReadBitsAvailable = pInst->iWriteBitIndex; // .. and only read as many bits as placed there by SBP_WriteBitgroups() in the PREVIOUS (odd) pass
     iBitgroupIndex = 0;
     while( pInst->iReadBitIndex < pInst->nReadBitsAvailable && ( iBitgroupIndex < MAX_BITGROUPS_PER_PACKET_BUFFER_TEST ) )
      {
        u32Value   = u32TestBitgroupValues[ iBitgroupIndex ];
        iBitgroupSize = iTestBitgroupSizes[ iBitgroupIndex ];

        // Limit the test value to match number of bits (otherwise the test fails
        //           even though SBP_ReadBitgroup() is ok):
        u32Value &= (0xFFFFFFFFUL >> (32 - iBitgroupSize) );
            // ,------------|____________________________________|
            // '--> Extreme case: If iBitgroupSize was ZERO,
            //      the entire 0xFFFFFFFF would be 'shifted out' to the left,
            //      and the only possible u32Value would be zero.
            // Of course that's nonsense; the minimum bitgroup size is ONE.
            //
        u32ValueRead = SBP_ReadBitgroup( pInst, iBitgroupSize );
           // AFTER calling SBP_ReadBitgroup(),
           //   pInst->iReadBitIndex has been incremented by the number
           //   of bits READ FROM the packet buffer, which includes the
           //   "cleared bit SEVENs" in each byte of the buffer !
           //   Thus, pInst->iReadBitIndex may step by MORE than iBitgroupSize.
           //   Because this test uses random data, we'll spot errors from an
           //   improperly incremented BIT INDICES quickly .
           //   (and indeed, in early versions, an old unit test filling packets
           //    with "all-ones" and "all-zeros" did NOT spot errors that were
           //    detected using bitgroups of pseudo-random length (1..32 bits),
           //    and pseudo-random values.)
           //
        if( u32Value != u32ValueRead )
         {
           ++nErrors;
#         if( 1 && SBP_COMPILING_SPECLAB )// only used during the early development phase:
           DEBUG_EnterErrorHistory( DEBUG_LEVEL_ERROR, 0, UTL_USE_CURRENT_TIME,
               "Error testing SevenBitPackets.c : Read %d-bit-group 0x%lX, expected 0x%lX",
               (int)iBitgroupSize, (DWORD)u32ValueRead, (DWORD)u32Value );
#         endif // SBP_COMPILING_SPECLAB ?
         }
        ++iBitgroupIndex;
      }    // end while( iBitgroupIndex ... ) to READ/VERIFY

   } // end for( iPass .. )

#if( SBP_USE_FLOATING_POINT ) // Thest the conversion from 'double' to our "fixed point format with 21-bit fractional part" ?
  d1 = -1048575.000001;
  SBP_DoubleToFixedPoint21( d1, &integer_part, &fractional_part );
  if( (integer_part != -1048575) || ( fractional_part != -1 ) )
   { ++nErrors;
#   if( 1 && SBP_COMPILING_SPECLAB )// only used during the early development phase:
     DEBUG_EnterErrorHistory( DEBUG_LEVEL_ERROR, 0, UTL_USE_CURRENT_TIME,
               "Error in SBP_DoubleToFixedPoint21(%.12lg) : got int=%ld, fract=%ld / 2^21",
               (double)d1, (long)integer_part, (long)fractional_part );
#   endif // SBP_COMPILING_SPECLAB ?
   }
  d2 = SBP_FixedPoint21ToDouble( integer_part, fractional_part ); // convert BACK..
  if( (d2 < (d1 - 1E-6) ) // (almost) the same as the original value ?
    ||(d2 > (d2 + 1E-6) ) // 20 bit fractional part (without the sign bit) should be ok for at least SIX decimal places
    )
   { ++nErrors;
#   if( 1 && SBP_COMPILING_SPECLAB )
     DEBUG_EnterErrorHistory( DEBUG_LEVEL_ERROR, 0, UTL_USE_CURRENT_TIME,
               "Error in SBP_FixedPoint21ToDouble() : got %.12lg, expected %.12lg .",
                (double)d2, (double)d1 );
#   endif // SBP_COMPILING_SPECLAB ?
   }

#endif // SBP_USE_FLOATING_POINT ?

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  // Above: Tests for the low-level "payload assembly / disassembly" functions.
  // Below: Tests with a complete packets; at least the most important ones.
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  for( iPass=0; iPass<=1; ++iPass ) // pass 0 : 'Audio sample format' with INTEGER sampling rate; pass 1 : .. with non-integer SR
   { int nBytes, nBytesExpected;
#   if(SBP_USE_FLOATING_POINT)
     double dblExpectedSampleRate;
#   endif // SBP_USE_FLOATING_POINT ?
     pInst->nBitsPerSingleSample    = 24;  // test the "densely packet audio sample format" (info packet) ..
     pInst->nChannelsPerSamplePoint = 4;
     pInst->iSampleDataType         = SBP_DATA_TYPE_SIGNED_LE; // use signed integer, Little endian.
#   if(SBP_USE_FLOATING_POINT)
     pInst->dblSampleRate_Hz = 48000.000;  // Samples per second, precise, possibly sent as a FIXED POINT VALUE (with a fractional part)
     nBytesExpected = 8;     // 2-byte header and at least 6-byte payload (with 7 bits usable per payload-byte)
     if( iPass==1 )
      { pInst->dblSampleRate_Hz = 48000.123456;
        nBytesExpected += 3; // expecting 3 more bytes for the 21-bit FRACTIONAL part
      }
     dblExpectedSampleRate = pInst->dblSampleRate_Hz; // for comparison further below
#   else  // for small microcontrollers without floating point support (or too little ROM) :
     pInst->i32SampleRate_Hz = 48000;  // "integer only" : boring, not tested further below
#   endif // SBP_USE_FLOATING_POINT ?

     nBytes = SBP_WritePacket_AudioSampleFormat( pInst ); // .. into pInst->bPacketBuffer[0..i-1]
     if( nBytes != nBytesExpected )  // with an INTEGER sampling rate, no need for a fractional part, thus expecting a total of EIGHT bytes (header+payload)
      { ++nErrors;
#      if( 1 && SBP_COMPILING_SPECLAB )// only used during the early development phase:
        DEBUG_EnterErrorHistory( DEBUG_LEVEL_ERROR, 0, UTL_USE_CURRENT_TIME,
           "Error testing SevenBitPackets.c : Bad 'Audio-Sample-Format' packet length (%d, expected: %d)",
           (int) nBytes, (int)nBytesExpected );
#      endif // SBP_COMPILING_SPECLAB ?
      }
     pInst->nBitsPerSingleSample = pInst->nChannelsPerSamplePoint = pInst->iSampleDataType = -1; // no cheating..
#   if(SBP_USE_FLOATING_POINT)
     pInst->dblSampleRate_Hz = -1;  // .. we want to extract ALL THIS from the AUDIO_SAMPLE_FORMAT assembled above !
#   endif // SBP_USE_FLOATING_POINT ?
     SBP_ProcessReceivedData( pInst, pInst->bPacketBuffer, i/*nBytesRcvd*/, NULL/*no callback*/, NULL/*no "user data"*/ );
     // If all worked as planned, the above packet is recognized as
     // SB_PACKET_CONTENT_TYPE_AUDIO_SAMPLE_FORMAT, and the 'deliberately trashed'
     // info in pInst is valid again. Check that info, extracted from the "received" packet:
     if(  ( pInst->nBitsPerSingleSample    != 24 )
        ||( pInst->nChannelsPerSamplePoint != 4  )
        ||( pInst->iSampleDataType         != SBP_DATA_TYPE_SIGNED_LE )
#      if( SBP_USE_FLOATING_POINT )
        ||( pInst->dblSampleRate_Hz < (dblExpectedSampleRate-1E-6) ) // the devil of 'limited resolution'..
        ||( pInst->dblSampleRate_Hz > (dblExpectedSampleRate+1E-6) )
            // (indeed, the test threw an error here without the above "tolerance window",
            //  which is plausible: 1 / 2^20 = 0.000000953, that's NOT ALWAYS sufficient
            //  for six decimal places, even with a 'double' for the internal calculation.
            //  Seen at this point: pInst->dblSampleRate_Hz = 48000.1234550478 instead of
            //                        dblExpectedSampleRate = 48000.1234560000 .
            //         Difference : 9.522e-7 .  1 / 2^20 = 9.5367e-7 .   Ok  !
#      endif // SBP_USE_FLOATING_POINT ?
        )
      { ++nErrors;
#      if( 1 && SBP_USE_FLOATING_POINT && SBP_COMPILING_SPECLAB )// only used during the early development phase:
        DEBUG_EnterErrorHistory( DEBUG_LEVEL_ERROR, 0, UTL_USE_CURRENT_TIME,
           "Error testing SevenBitPackets.c : Bad data in decoded 'Audio-Sample-Format': %d bits, %d chns, %.12lg Samp/sec; expected %.12lg",
           (int)pInst->nBitsPerSingleSample, (int)pInst->nChannelsPerSamplePoint, (double)pInst->dblSampleRate_Hz, (double)dblExpectedSampleRate );
#      endif // SBP_COMPILING_SPECLAB ?
      }
   } // end for( iPass .. )


  return nErrors;
} // end SBP_RunUnitTests()
#endif // SBP_BUILT_IN_UNIT_TESTS ?


// EOF < SevenBitPackets.c > . Leave an empty line below, for stupid compilers.

